@@ -1,4 +1,4 @@
-import select, socket, threading, traceback, sys, time
+import select, socket, threading, sys, time, traceback
 from OpenSSL import SSL
 import certutils
 
@@ -35,8 +35,8 @@ OP_TLS_ROLLBACK_BUG = SSL.OP_TLS_ROLLBACK_BUG
 
 
 class NetLibError(Exception): pass
-class NetLibDisconnect(Exception): pass
-class NetLibTimeout(Exception): pass
+class NetLibDisconnect(NetLibError): pass
+class NetLibTimeout(NetLibError): pass
 
 
 class _FileLike:
@@ -44,6 +44,7 @@ class _FileLike:
     def __init__(self, o):
         self.o = o
         self._log = None
+        self.first_byte_timestamp = None
 
     def set_descriptor(self, o):
         self.o = o
@@ -80,16 +81,25 @@ class _FileLike:
         if self.is_logging():
             self._log.append(v)
 
+    def reset_timestamps(self):
+        self.first_byte_timestamp = None
+
 
 class Writer(_FileLike):
     def flush(self):
-        try:
-            if hasattr(self.o, "flush"):
+        """
+            May raise NetLibDisconnect
+        """
+        if hasattr(self.o, "flush"):
+            try:
                 self.o.flush()
-        except socket.error, v:
-            raise NetLibDisconnect(str(v))
+            except socket.error, v:
+                raise NetLibDisconnect(str(v))
 
     def write(self, v):
+        """
+            May raise NetLibDisconnect
+        """
         if v:
             try:
                 if hasattr(self.o, "sendall"):
@@ -129,8 +139,9 @@ class Reader(_FileLike):
                 raise NetLibTimeout
             except socket.error:
                 raise NetLibDisconnect
-            except SSL.SysCallError, v:
+            except SSL.SysCallError:
                 raise NetLibDisconnect
+            self.first_byte_timestamp = self.first_byte_timestamp or time.time()
             if not data:
                 break
             result += data
@@ -168,12 +179,19 @@ class TCPClient:
         self.cert = None
         self.ssl_established = False
 
-    def convert_to_ssl(self, clientcert=None, sni=None, method=TLSv1_METHOD, options=None):
+    def convert_to_ssl(self, cert=None, sni=None, method=TLSv1_METHOD, options=None):
+        """
+            cert: Path to a file containing both client cert and private key.
+        """
         context = SSL.Context(method)
-        if not options is None:
-            ctx.set_options(options)
-        if clientcert:
-            context.use_certificate_file(self.clientcert)
+        if options is not None:
+            context.set_options(options)
+        if cert:
+            try:
+                context.use_privatekey_file(cert)
+                context.use_certificate_file(cert)
+            except SSL.Error, v:
+                raise NetLibError("SSL client certificate error: %s"%str(v))
         self.connection = SSL.Connection(context, self.connection)
         self.ssl_established = True
         if sni:
@@ -202,7 +220,7 @@ class TCPClient:
         self.connection.settimeout(n)
 
     def gettimeout(self):
-        self.connection.gettimeout()
+        return self.connection.gettimeout()
 
     def close(self):
         """
@@ -222,6 +240,7 @@ class TCPClient:
 class BaseHandler:
     """
         The instantiator is expected to call the handle() and finish() methods.
+            
     """
     rbufsize = -1
     wbufsize = -1
@@ -235,20 +254,47 @@ class BaseHandler:
         self.finished = False
         self.ssl_established = False
 
-    def convert_to_ssl(self, cert, key, method=SSLv23_METHOD, options=None):
+        self.clientcert = None
+
+    def convert_to_ssl(self, cert, key, method=SSLv23_METHOD, options=None, handle_sni=None, request_client_cert=False):
         """
             method: One of SSLv2_METHOD, SSLv3_METHOD, SSLv23_METHOD, or TLSv1_METHOD
+            handle_sni: SNI handler, should take a connection object. Server
+            name can be retrieved like this:
+
+                            connection.get_servername()
+
+                        And you can specify the connection keys as follows:
+
+                            new_context = Context(TLSv1_METHOD)
+                            new_context.use_privatekey(key)
+                            new_context.use_certificate(cert)
+                            connection.set_context(new_context)
+
+            The request_client_cert argument requires some explanation. We're
+            supposed to be able to do this with no negative effects - if the
+            client has no cert to present, we're notified and proceed as usual.
+            Unfortunately, Android seems to have a bug (tested on 4.2.2) - when
+            an Android client is asked to present a certificate it does not
+            have, it hangs up, which is frankly bogus. Some time down the track
+            we may be able to make the proper behaviour the default again, but
+            until then we're conservative.
         """
         ctx = SSL.Context(method)
         if not options is None:
             ctx.set_options(options)
-        ctx.set_tlsext_servername_callback(self.handle_sni)
+        if handle_sni:
+            # SNI callback happens during do_handshake()
+            ctx.set_tlsext_servername_callback(handle_sni)
         ctx.use_privatekey_file(key)
         ctx.use_certificate_file(cert)
+        if request_client_cert:
+            def ver(*args):
+                self.clientcert = certutils.SSLCert(args[1])
+            ctx.set_verify(SSL.VERIFY_PEER, ver)
         self.connection = SSL.Connection(ctx, self.connection)
         self.ssl_established = True
         self.connection.set_accept_state()
-        # SNI callback happens during do_handshake()
         try:
             self.connection.do_handshake()
         except SSL.Error, v:
@@ -264,27 +310,10 @@ class BaseHandler:
             self.close()
             self.wfile.close()
             self.rfile.close()
-        except socket.error:
+        except (socket.error, NetLibDisconnect):
             # Remote has disconnected
             pass
 
-    def handle_sni(self, connection):
-        """
-            Called if the client has given a server name indication.
-
-            Server name can be retrieved like this:
-
-                connection.get_servername()
-
-            And you can specify the connection keys as follows:
-
-                new_context = Context(TLSv1_METHOD)
-                new_context.use_privatekey(key)
-                new_context.use_certificate(cert)
-                connection.set_context(new_context)
-        """
-        pass
-
     def handle(self): # pragma: no cover
         raise NotImplementedError
 
@@ -300,7 +329,7 @@ class BaseHandler:
                 self.connection.shutdown()
             else:
                 self.connection.shutdown(socket.SHUT_RDWR)
-        except (socket.error, SSL.Error), v:
+        except (socket.error, SSL.Error):
             # Socket probably already closed
             pass
         self.connection.close()
@@ -333,20 +362,13 @@ class TCPServer:
             while not self.__shutdown_request:
                 r, w, e = select.select([self.socket], [], [], poll_interval)
                 if self.socket in r:
-                    try:
-                        request, client_address = self.socket.accept()
-                    except socket.error:
-                        return
-                    try:
-                        t = threading.Thread(
-                                target = self.request_thread,
-                                args = (request, client_address)
-                            )
-                        t.setDaemon(1)
-                        t.start()
-                    except:
-                        self.handle_error(request, client_address)
-                        request.close()
+                    request, client_address = self.socket.accept()
+                    t = threading.Thread(
+                            target = self.request_thread,
+                            args = (request, client_address)
+                        )
+                    t.setDaemon(1)
+                    t.start()
         finally:
             self.__shutdown_request = False
             self.__is_shut_down.set()
